from lattice_data import Lattice
from lattice_synthesis import VivdoHLS_Synthesis
from lattice_synthesis import FakeSynthesis
from lattice_ds_point import DSpoint
from lattice_sphere_tree import SphereTree as st
import lattice_utils
import datasets
import numpy as np
from itertools import zip_longest
import copy
from matplotlib import pyplot as plt
import sys

sys.setrecursionlimit(15200)
# sys.setrecursionlimit(10000)
np.random.seed(42)

# Set the radius for the exploration
radius = 3

# Set number of runs due to the probabilistic initial sampling
n_of_runs = 2

# To plot the ADRS chart
plot = True

# Initial sampling size
intial_sampling_size = 10

# Performance goal
pf_goal = 0

# Collect stats
max_n_of_synth = 0
adrs_run_stats = []
goal_stats_history = []
goal_stats = []

# Read dataset

# Specify the list of dataset to explore
# benchmark = ["ChenIDCt", "adpcm_decode", "adpcm_encode", "Autocorrelation", "Reflection_coefficients"]
benchmark = ["ChenIDCt"]
for b in benchmark:
    # Extract data from the database
    database = datasets.Datasets(b)
    synthesis_result = database.benchmark_synthesis_results
    configurations = database.benchmark_configurations
    feature_sets = database.benchmark_feature_sets
    directives = database.benchmark_directives

    # Format data for the exploration
    entire_ds = []
    for i in range(0, len(synthesis_result)):
        entire_ds.append(DSpoint(synthesis_result[i][0], synthesis_result[i][1], list(configurations[i])))

    # Set variables to store exploration data
    adrs_run_history = []
    run_stats = []

    for run in range(0, n_of_runs):
        goal_acheived = False
        # Collect stats
        online_statistics = {}
        online_statistics['adrs'] = []
        online_statistics['delta_adrs'] = []
        online_statistics['n_synthesis'] = []
        # run_stats = None

        print("Exploration n: " + str(run))
        sphere_elements_sizes = []
        max_radius = 0
        # Create Lattice
        lattice = Lattice(feature_sets, radius)

        # Generate inital samples
        samples = lattice.beta_sampling(0.1, 0.1, intial_sampling_size)

        # samples = samples_dataset
        n_of_synthesis = len(samples)

        # Synthesise sampled configuration
        # FakeSynthesis simulates the synthesis process retrieving the configuration from the proper DB
        hls = FakeSynthesis(entire_ds, lattice)
        sampled_configurations_synthesised = []
        for s in samples:
            latency, area = hls.synthesise_configuration(s)
            synthesised_configuration = DSpoint(latency, area, s)
            sampled_configurations_synthesised.append(synthesised_configuration)
            lattice.lattice.add_config(s)

        # After the inital sampling, retrieve the pareto frontier
        pareto_frontier, pareto_frontier_idx = lattice_utils.pareto_frontier2d(sampled_configurations_synthesised)

        # Get exhaustive pareto frontier (known only if ground truth exists)
        pareto_frontier_exhaustive, pareto_frontier_exhaustive_idx = lattice_utils.pareto_frontier2d(entire_ds)

        # Store a copy to save the pareto front before the exploration algorithm
        pareto_frontier_before_exploration = copy.deepcopy(pareto_frontier)

        # Calculate ADRS
        adrs_evolution = []
        adrs = lattice_utils.adrs2d(pareto_frontier_exhaustive, pareto_frontier)
        # ADRS after initial sampling
        adrs_evolution.append(adrs)

        # Select randomly a pareto configuration and explore its neighbourhood
        r = np.random.randint(0, len(pareto_frontier))
        pareto_configurations = [samples[i] for i in pareto_frontier_idx]
        pareto_solution_to_explore = pareto_configurations[r]

        # Search locally for the configuration to explore
        sphere = st(pareto_solution_to_explore, lattice)
        new_configuration = sphere.random_closest_element

        # Until there are configurations to explore, try to explore these
        while new_configuration is not None:
            # Synthesise configuration
            latency, area = hls.synthesise_configuration(new_configuration)
            # Generate a new design point
            ds_point = DSpoint(latency, area, new_configuration)

            # Update known synthesis values and configurations(only pareto + the new one)
            pareto_frontier.append(ds_point)

            # Add configuration to the tree
            lattice.lattice.add_config(ds_point.configuration)

            # Get pareto frontier
            pareto_frontier, pareto_frontier_idx = lattice_utils.pareto_frontier2d(pareto_frontier)

            # Calculate ADRS
            adrs = lattice_utils.adrs2d(pareto_frontier_exhaustive, pareto_frontier)
            adrs_evolution.append(adrs)

            # Find new configuration to explore
            # Select randomly a pareto configuration
            search_among_pareto = copy.copy(pareto_frontier)
            while len(search_among_pareto) > 0:
                r = np.random.randint(0, len(search_among_pareto))
                pareto_solution_to_explore = search_among_pareto[r].configuration

                # Explore the closer element locally
                sphere = st(pareto_solution_to_explore, lattice)
                new_configuration = sphere.random_closest_element
                if new_configuration is None:
                    search_among_pareto.pop(r)
                    continue

                max_radius = max(max_radius, sphere.radius)
                if max_radius > lattice.max_distance:
                    search_among_pareto.pop(r)
                    continue

                break

            exit_expl = False
            if len(search_among_pareto) == 0:
                print("Exploration terminated")
                exit_expl = True

            if max_radius > lattice.max_distance:
                print("Max radius reached")
                exit_expl = True

            # Here eventually add a condition to limit the number of synthesis to a certain treshold
            # if n_of_syntheis == budget:
            #     exit_expl = True

            if adrs <= pf_goal and not goal_acheived:
                goal_acheived = True
                goal_stats.append(n_of_synthesis)
                exit_expl = True

            if max_n_of_synth < n_of_synthesis:
                max_n_of_synth = n_of_synthesis

            n_of_synthesis += 1
            run_stats = lattice_utils.collect_online_statis(online_statistics, adrs, n_of_synthesis)

            if exit_expl:
                # If the exploration is ending update the calculate final ADRS
                adrs = lattice_utils.adrs2d(pareto_frontier_exhaustive, pareto_frontier)
                adrs_evolution.append(adrs)
                print("Number of synthesis:\t{:d}".format(n_of_synthesis))
                print("Max radius:\t{:0.4f}".format(max_radius))
                print("Final ADRS:\t{:0.4f}".format(adrs))
                print()
                # print(lattice.lattice.get_n_of_children())
                # goal_stats.append(n_of_synthesis)
                break


        adrs_run_stats.append(run_stats)
        adrs_run_history.append(adrs_evolution)
        # goal_stats_history.append(goal_stats)

collect_offline_stats = lattice_utils.collect_offline_stats(adrs_run_stats, n_of_runs, max_n_of_synth, goal_stats)

# print "final_adrs"
# print collect_offline_stats['final_adrs']
# print "final_adrs_outliers"
# print collect_offline_stats['final_adrs_outliers']
# print "adrs_mean"
# print collect_offline_stats['adrs_mean']
# print "adrs"
# print collect_offline_stats['adrs']
# print "delta_adrs"
# print collect_offline_stats['delta_adrs']
# print "delta_adrs_mean"
# print collect_offline_stats['delta_adrs_mean']
# print "adrs_mean_outliers"
# print collect_offline_stats['adrs_mean_outliers']
# print "delta_adrs_mean_outliers"
# print collect_offline_stats['delta_adrs_mean_outliers']
# print "max_n_of_synth"
# print collect_offline_stats['max_n_of_synth']
# print "max_n_of_synth_mean"
# print collect_offline_stats['max_n_of_synth_mean']
# print "max_n_of_synth_mean_outliers"
# print collect_offline_stats['max_n_of_synth_mean_outliers']

print (goal_stats)

# Plot ADRS evolution
if plot:
    averages_adrs = list(list(map(lattice_utils.avg, zip_longest(*adrs_run_history))))
    plt.title("ADRS evolution")
    plt.ylabel("mean ADRS")
    plt.xlabel("# of synthesis")
    plt.plot(range(intial_sampling_size,len(averages_adrs)+intial_sampling_size), averages_adrs)
    plt.grid()
    # plt.show()
    plt.savefig('ADRS evolution.png')

    fig1, ax1 = plt.subplots()
    ax1.set_title('max_n_of_synth')
    # ax1.boxplot(collect_offline_stats['final_adrs'])
    ax1.boxplot(collect_offline_stats['max_n_of_synth'])
    # plt.show()
    plt.savefig('max_n_of_synth.png')

    fig2, ax2 = plt.subplots()
    ax2.set_title('final_adrs')
    ax2.boxplot(collect_offline_stats['final_adrs'])
    # ax2.boxplot(collect_offline_stats['max_n_of_synth'])
    # plt.show()
    plt.savefig('final_adrs.png')

    fig3, ax3 = plt.subplots()
    plt.title("ADRS evolution")
    plt.ylabel("ADRS")
    plt.xlabel("# of synthesis")
    for run in collect_offline_stats['adrs']:
        plt.plot(range(intial_sampling_size, len(run) + intial_sampling_size), run)
    plt.grid()
    # plt.show()
    plt.savefig('ADRS evolution all.png')

    fig4, ax4 = plt.subplots()
    plt.title("delta ADRS evolution")
    plt.ylabel("delta ADRS")
    plt.xlabel("# of synthesis")
    for run in collect_offline_stats['delta_adrs']:
        plt.plot(range(intial_sampling_size, len(run) + intial_sampling_size), run)
        # print run
    plt.grid()
    plt.savefig('delta ADRS evolution all.png')
    # plt.show()

    fig5, ax5 = plt.subplots()
    plt.title("Cumulative Dist")
    plt.ylabel("Probability")
    plt.xlabel("# of synthesis")
    goal_stats = np.asarray(goal_stats)
    bins = np.arange(np.floor(goal_stats.min()), goal_stats.max())
    values, base = np.histogram(goal_stats, bins=bins)
    # values, base = np.histogram(goal_stats, normed=True)
    # evaluate the cumulative
    cumulative = np.cumsum(values)
    # plot the cumulative function
    plt.plot(base[:-1], cumulative, c='blue')
    plt.grid()
    # plt.show()
    plt.savefig('Cumulative Dist.png')

    # fig5, ax5 = plt.subplots()
    # ax5.set_title('Boxplot goal')
    # tmp = []
    # for run in collect_offline_stats['adrs']:
    #     print run
    #     for i in xrange(0, len(run)):
    #         if run[i] <= pf_goal:
    #             tmp.append(intial_sampling_size+i)
    #             break
    # ax5.boxplot(tmp)
    # # ax2.boxplot(collect_offline_stats['max_n_of_synth'])
    # # plt.show()
    # plt.savefig('boxplot goal.png')



